from typing_extensions import TypedDict, Coroutine
import asyncio
from kv import KV
from haskellian import either as E
from pipeteer import Task
from dslog import Logger
import moveread.pipelines.preprocess as pre
from ._types import Input, Game

class Params(TypedDict):
  games: KV[Game]
  imgGameIds: KV[str]
  logger: Logger

class Preinput(Task[Input, pre.Input, Params, Coroutine]):
  Queues = Task.Queues[Input, pre.Input]
  Params = Params
  Artifacts = Coroutine

  def __init__(self):
    super().__init__(Input, pre.Input)

  async def run(self, queues: Queues, params: Params):
    Qin, Qout = queues['Qin'], queues['Qout']
    games, imgGameIds, logger = params['games'], params['imgGameIds'], params['logger']
      
    @E.do()
    async def input_one():
      gameId, task = (await Qin.read()).unsafe()
      logger(f'Processing "{gameId}"', level='DEBUG')
      imgIds = [f'{gameId}/{i}' for i in range(len(task.imgs))]
      (await games.insert(gameId, Game(model=task.model, imgIds=imgIds))).unsafe()
      E.sequence(await asyncio.gather(*[
        imgGameIds.insert(imgId, gameId)
        for imgId in imgIds
      ])).unsafe()
      E.sequence(await asyncio.gather(*[
        Qout.push(imgId, pre.Input.new(model=task.model, img=img))
        for imgId, img in zip(imgIds, task.imgs)
      ])).unsafe()
      (await Qin.pop(gameId)).unsafe()
      logger(f'Pushed tasks from "{gameId}"', level='DEBUG')

    while True:
      try:
        r = await input_one()
        if r.tag == 'left':
          logger(r.value, level='ERROR')
          await asyncio.sleep(1)
      except Exception as e:
        logger('Unexpected exception:', e, level='ERROR')
        await asyncio.sleep(1)



from typing import Callable
from uuid import uuid4
from haskellian import either as E
from kv import KV
from pipeteer import WriteQueue
from dslog import Logger
import robust_extraction2 as re
from moveread.core import Core, Game
from moveread.pipelines.game_preprocess import Input

async def input_core(
  core: Core, Qin: WriteQueue[Input],
  *, images: KV[bytes],
  model_fn: Callable[[Game], re.ExtendedModel],
  num_games: int | None = None, shuffle: bool = True,
  logger = Logger.rich().prefix('[CORE INPUT]')
):
  """Input all images from `core` into `Qin` tasks
  - Actually, only images with `version == 0`
  - `model_fn`: determines the scoresheet model of each task
  - `state_fn`: determines an arbitrary tuple of JSON-serializable data to attach to each task
  """
  games = await core.games.keys().map(E.unsafe).sync()
  if shuffle:
    import random
    random.shuffle(games)
  for gameId in games[:num_games]:
    game = (await core.games.read(gameId)).unsafe()
    imgs = []
    for _, image in game.images:
      url = f'{gameId}/original_{uuid4()}.jpg'
      img = (await core.blobs.read(image.url)).unsafe()
      (await images.insert(url, img)).unsafe()
      imgs.append(url)
      break

    task = Input(model=model_fn(game), imgs=imgs)
    await Qin.push(gameId, task)
    logger(f'Inputted task "{gameId}". Task:', task)


from argparse import ArgumentParser

def main():
  parser = ArgumentParser()
  parser.add_argument('-d', '--db-path', required=True)
  parser.add_argument('--images', required=True)
  parser.add_argument('--url', required=True, help='The server URL')

  parser.add_argument('-p', '--port', default=8000, type=int)
  parser.add_argument('--host', default='0.0.0.0', type=str)

  args = parser.parse_args()


  import os
  from dslog import Logger
  db_path = os.path.join(os.getcwd(), args.db_path)
  
  logger = Logger.click().prefix('[GAME PREPROCESS]')
  logger(f'Running preprocessing...')
  logger(f'Images conn str: "{args.images}"')
  logger(f'DB path: "{db_path}"')
  
  import asyncio
  from multiprocessing import Process
  from kv import KV, SQLiteKV, LocatableKV, ServerKV
  from pipeteer import QueueKV
  import uvicorn
  from fastapi.middleware.cors import CORSMiddleware
  import moveread.pipelines.preprocess as pre
  from moveread.pipelines.game_preprocess import GamePreprocess, Output, Game

  images = KV[bytes].of(args.images)
  if not isinstance(images, LocatableKV):
    if not args.url:
      raise ValueError('Provide a LocatableKV (--images) or a base URL (--url)')
    images = images.served(args.url.rstrip('/') + '/images')

  def get_queue(path, type: type):
    return QueueKV.sqlite(type, db_path, table='-'.join(['queues', *path]))
  
  Qout = get_queue(('output',), Output)
  wkf = GamePreprocess()

  params = GamePreprocess.Params(
    logger=logger, images=images,
    buffer=SQLiteKV.at(db_path, dict[str, pre.Output], table='buffer'),
    games=SQLiteKV.at(db_path, Game, table='games'),
    imgGameIds=SQLiteKV.at(db_path, str, table='game-ids'),
    descaled_h=512
  )

  Qs = wkf.connect(Qout, get_queue, params)
  artifs = wkf.run(Qs, params)
  artifs.api.add_middleware(CORSMiddleware, allow_origins=['*'], allow_credentials=True, allow_methods=['*'], allow_headers=['*'])
  artifs.api.mount('/images', ServerKV(images))

  ps = {
    id: Process(target=asyncio.run, args=(f,)) for id, f in artifs.processes.items()
  } | {
    'api': Process(target=uvicorn.run, args=(artifs.api,), kwargs={'host': args.host, 'port': args.port})
  }
  for id, p in ps.items():
    p.start()
    logger(f'Process "{id}" started at PID {p.pid}')
  for p in ps.values():
    p.join()
  

if __name__ == '__main__':
  import sys
  import os
  os.chdir('/home/m4rs/mr-github/rnd/data/moveread-pipelines/backend/3.game-preprocess/')
  sys.argv.extend('-q demo/queues.sqlite --images demo/images/'.split(' '))
  main()


from typing_extensions import TypedDict, Coroutine
import asyncio
from kv import KV
from haskellian import either as E
from pipeteer import Task
from dslog import Logger
import moveread.pipelines.preprocess as pre
from ._types import Game, Output

class Params(TypedDict):
  buffer: KV[dict[str, pre.Output]]
  games: KV[Game]
  imgGameIds: KV[str]
  logger: Logger

class Join(Task[pre.Output, Output, Params, Coroutine]):
  Queues = Task.Queues[pre.Output, Output]
  Params = Params
  Artifacts = Coroutine

  def __init__(self):
    super().__init__(pre.Output, Output)

  async def run(self, queues: Queues, params: Params):
    Qin, Qout = queues['Qin'], queues['Qout']
    buffer, games, imgGameIds, logger = params['buffer'], params['games'], params['imgGameIds'], params['logger']
      
    @E.do()
    async def run_one():
      imgId, result = (await Qin.read()).unsafe()
      logger(f'Processing "{imgId}"', level='DEBUG')
      gameId = (await imgGameIds.read(imgId)).mapl(lambda err: f'Error reading image "{imgId}" gameId: {err}').unsafe()
      logger(f'Received "{imgId}" for "{gameId}"', level='DEBUG')
      game, received = await asyncio.gather(
        games.read(gameId).then(lambda e: e.mapl(lambda err: f'Error reading buffered game: {err}').unsafe()),
        buffer.read(gameId).then(E.get_or({})),
      )

      received_now = dict(received) | { imgId: result }
      receivedIds = set(imgId for imgId, _ in received_now.items())
      requiredIds = set(game.imgIds)
      logger('Received:', receivedIds, 'Required', requiredIds, level='DEBUG')

      if receivedIds == requiredIds:
        next = [received_now[id] for id in game.imgIds]
        (await Qout.push(gameId, next)).unsafe()
        _, e = await asyncio.gather(
          games.delete(gameId).then(E.unsafe),
          buffer.delete(gameId),
        )
        if e.tag == 'left' and e.value.reason != 'inexistent-item':
          e.unsafe()
        logger(f'Joined results for {gameId}')
      else:
        (await buffer.insert(gameId, received_now)).unsafe()
      
      (await imgGameIds.delete(imgId)).unsafe()
      (await Qin.pop(imgId)).unsafe()

    while True:
      try:
        r = await run_one()
        if r.tag == 'left':
          logger(r.value, level='ERROR')
          await asyncio.sleep(1)
      except Exception as e:
        logger('Unexpected exception:', e, level='ERROR')
        await asyncio.sleep(1)
        


from typing import Sequence
from dataclasses import dataclass
import robust_extraction2 as re
import moveread.pipelines.preprocess as pre

@dataclass
class Input:
  model: re.ExtendedModel
  imgs: Sequence[str]

@dataclass
class Game:
  model: re.ExtendedModel
  imgIds: Sequence[str]

Output = Sequence[pre.Output]


from typing_extensions import TypedDict, NotRequired
from kv import KV, LocatableKV
from dslog import Logger
from pipeteer import Workflow
import moveread.pipelines.preprocess as pre
from ._types import Input, Output, Game
from .join import Join
from .preinput import Preinput

class Pipelines(TypedDict):
  preinput: Preinput
  preprocess: pre.Preprocess
  join: Join

class Queues(TypedDict):
  preinput: Preinput.Queues
  preprocess: pre.Preprocess.Queues
  join: Join.Queues

class StorageParams(TypedDict):
  images: LocatableKV[bytes]
  buffer: KV[dict[str, pre.Output]]
  games: KV[Game]
  imgGameIds: KV[str]

class Params(StorageParams):
  logger: NotRequired[Logger]
  descaled_h: int
  model_path: str

class GamePreprocess(Workflow[Input, Output, Queues, Params, pre.Artifacts, Pipelines]): # type: ignore
  Input = Input
  Output = Output
  Queues = Queues
  Params = Params
  Artifacts = pre.Artifacts

  def __init__(self):
    super().__init__({
      'preinput': Preinput(),
      'preprocess': pre.Preprocess(),
      'join': Join(),
    })

  def run(self, queues: Queues, params: Params):

    logger = params.get('logger') or Logger.click(); images = params['images']
    buffer = params['buffer']; games = params['games']; imgGameIds = params['imgGameIds']

    artifs = self.pipelines['preprocess'].run(queues['preprocess'], pre.Params(logger=logger, images=images, descaled_h=params['descaled_h'], model_path=params['model_path']))
    artifs.processes = {
      f'preprocess-{k}': v for k, v in artifs.processes.items()
    } | {
      'preinput': self.pipelines['preinput'].run(queues['preinput'], Preinput.Params(logger=logger, games=games, imgGameIds=imgGameIds)),
      'join': self.pipelines['join'].run(queues['join'], Join.Params(logger=logger, buffer=buffer, games=games, imgGameIds=imgGameIds)),
    }
    return artifs


from ._types import Input, Output, Game
from .spec import GamePreprocess, Params, StorageParams
from .core import input_core

__all__ = [
  'Input', 'Output', 'Game', 'GamePreprocess', 'input_core', 'Params', 'StorageParams',
]